from pathlib import Path
from unittest import TestCase

import mesido._darcy_weisbach as darcy_weisbach
from mesido.electricity_physics_mixin import ElectrolyzerOption
from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.network_common import NetworkSettings
from mesido.workflows.multicommodity_simulator_workflow import (
    MultiCommoditySimulator,
    MultiCommoditySimulatorNoLosses,
    run_sequatially_staged_simulation,
)

import numpy as np

from rtctools.util import run_optimization_problem

from utils_test_scaling import create_problem_with_debug_info, problem_scaling_check

from utils_tests import (
    demand_matching_test,
    electric_power_conservation_test,
    energy_conservation_test,
    feasibility_test,
)


def check_electrolyzer_efficiency(tol, electrolyzer_gas, electrolyzer_power, esdl_electrolyzer):
    efficiency = electrolyzer_power / electrolyzer_gas  # W/(g/s) = Ws/g
    efficiency = efficiency / 3600  # Wh/g
    provided_efficiencies = (
        esdl_electrolyzer.attributes["effMaxLoad"],
        esdl_electrolyzer.attributes["effMinLoad"],
        esdl_electrolyzer.attributes["efficiency"],
    )
    np.testing.assert_array_less(min(provided_efficiencies), efficiency[electrolyzer_power > 1])
    np.testing.assert_array_less(
        efficiency[electrolyzer_power > 0], max(provided_efficiencies) + tol
    )


def checks_all_mc_simulations(solution, results):
    # General checks
    demand_matching_test(solution, results)
    energy_conservation_test(solution, results)

    for prod in solution.energy_system_components.get("electricity_source", []):
        prod_profile_name = f"{prod}.maximum_electricity_source"
        energy_prod = results[f"{prod}.Electricity_source"]
        if prod_profile_name in solution.io.get_timeseries_names():
            target = solution.get_timeseries(prod_profile_name).values[: len(energy_prod)]
            np.testing.assert_allclose(target, energy_prod, atol=1.0e-3, rtol=1.0e-6)
        # print(prod, energy_prod)

    for demand in solution.energy_system_components.get("electricity_demand", []):
        energy_demand = results[f"{demand}.Electricity_demand"]
        if f"{demand}.target_electricity_demand" in solution.io.get_timeseries_names():
            target = solution.get_timeseries(f"{demand}.target_electricity_demand").values
            np.testing.assert_allclose(target, energy_demand, atol=1.0e-3, rtol=1.0e-6)
        #     print(demand, target, energy_demand)
        # else:
        #     print(demand, energy_demand)

    for demand in solution.energy_system_components.get("gas_demand", []):
        energy_demand = results[f"{demand}.Gas_demand_mass_flow"]
        if f"{demand}.target_gas_demand" in solution.io.get_timeseries_names():
            target = solution.get_timeseries(f"{demand}.target_gas_demand").values
            np.testing.assert_allclose(target, energy_demand, atol=1.0e-3, rtol=1.0e-6)
        #     print(demand, target, energy_demand)
        # else:
        #     print(demand, energy_demand)


class TestMultiCommoditySimulator(TestCase):
    # TODO: update docstring
    """
    In this test case several producers and consumers as well as conversion assets are applied, for
    the two commodities Hydrogen and electricity.
    The priority of the consumers, producers and conversion assets is set using the marginal costs
    in the ESDL file, allowing for flexible customised operation.

    Checks:
    - General checks namely demand matching, energy conservation and asset milp variable vs
      calculated milp (based on flow rate)
    - Check that producer 1 (merit oder = 2) is only used for the supply of milp lossed in the
      connected and is does not contribute to the heating demands 1, 2 and 3
    - Check that the ATES is not delivering any milp to the network during the 1st time step
    """

    def test_multi_commodity_simulator_priorities_el(self):
        import models.unit_cases_electricity.bus_networks.src.example as example

        base_folder = Path(example.__file__).resolve().parent.parent

        multicommoditysimulatorscaling, logger, logs_list = create_problem_with_debug_info(
            MultiCommoditySimulator
        )

        solution = run_optimization_problem(
            multicommoditysimulatorscaling,
            base_folder=base_folder,
            esdl_file_name="Electric_bus4_priorities.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_2.csv",
        )

        problem_scaling_check(logs_list, logger)
        bounds = solution.bounds()
        results = solution.extract_results()

        checks_all_mc_simulations(solution, results)

        prod_1 = results["ElectricityProducer_a215.Electricity_source"]
        prod_2 = results["ElectricityProducer_17a1.Electricity_source"]
        dem_1 = results["ElectricityDemand_e527.Electricity_demand"]
        dem_2 = results["ElectricityDemand_281a.Electricity_demand"]

        # check producer with highest priority (lowest marginal costs is maximizing production)
        np.testing.assert_allclose(
            prod_1, bounds["ElectricityProducer_a215.Electricity_source"][1], atol=1e-3, rtol=1e-6
        )
        # check producer with second highest priority is only producing to meet profile of demand
        prod_2_target = dem_1 - prod_1
        prod_2_target[prod_2_target < 0] = 0
        np.testing.assert_allclose(prod_2, prod_2_target, atol=1e-3, rtol=1e-6)
        # check demand with lowest marginal cost is only consuming if electricity left from producer
        # with highest priority after the matching of demand profile
        demand_2_target = prod_1 - dem_1
        demand_2_target[demand_2_target < 0] = 0
        np.testing.assert_allclose(dem_2, demand_2_target, atol=1e-3, rtol=1e-6)

    def test_multi_commodity_simulator_prod_profile(self):
        import models.unit_cases_electricity.bus_networks.src.example as example

        base_folder = Path(example.__file__).resolve().parent.parent

        solution = run_optimization_problem(
            MultiCommoditySimulator,
            base_folder=base_folder,
            esdl_file_name="Electric_bus4_priorities.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_2_prod.csv",
        )

        bounds = solution.bounds()
        results = solution.extract_results()

        checks_all_mc_simulations(solution, results)

        prod_1 = results["ElectricityProducer_a215.Electricity_source"]
        prod_2 = results["ElectricityProducer_17a1.Electricity_source"]
        dem_1 = results["ElectricityDemand_e527.Electricity_demand"]
        dem_2 = results["ElectricityDemand_281a.Electricity_demand"]

        # check producer with highest priority (lowest marginal costs is maximizing production) is
        # producing at max capacity, except when demand profile + max demand of other demand is
        # lower than prod_2 profile and max capacity prod_1
        np.testing.assert_allclose(
            prod_1[1:],
            bounds["ElectricityProducer_a215.Electricity_source"][1],
            atol=1e-3,
            rtol=1e-6,
        )
        np.testing.assert_allclose(
            prod_1[0],
            bounds["ElectricityDemand_281a.Electricity_demand"][1] + dem_1[0] - prod_2[0],
            atol=1e-3,
            rtol=1e-6,
        )

        # check demand with lowest marginal cost is only consuming if electricity left from producer
        # with highest priority and producer with profile, after the matching of demand profile
        demand_2_target = prod_1 + prod_2 - dem_1
        demand_2_target[demand_2_target < 0] = 0
        np.testing.assert_allclose(dem_2, demand_2_target, atol=1e-3, rtol=1e-6)

        # Check that producer 1 (merit oder = 2) is not used
        # and is does not contribute to the heating demands 1, 2 and 3
        # TODO: these tests need to be updated for elec/gas

        import models.unit_cases_gas.multi_demand_source_node.src.run_test as example

        base_folder = Path(example.__file__).resolve().parent.parent

        solution = run_optimization_problem(
            MultiCommoditySimulator,
            base_folder=base_folder,
            esdl_file_name="test_priorities.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries.csv",
        )

        results = solution.extract_results()
        bounds = solution.bounds()

        checks_all_mc_simulations(solution, results)

        prod_1 = results["GasProducer_3573.Gas_source_mass_flow"]
        prod_2 = results["GasProducer_a977.Gas_source_mass_flow"]
        dem_1 = results["GasDemand_47d0.Gas_demand_mass_flow"]
        dem_2 = results["GasDemand_7979.Gas_demand_mass_flow"]
        prod_1_bound = bounds["GasProducer_3573.Gas_source_mass_flow"][1]
        dem_2_bound = bounds["GasDemand_7979.Gas_demand_mass_flow"]

        # check producer with highest marginal cost (prod_2) is only producing to match demand
        # profile of dem_1 + the max of dem_2,
        # check producer_1 is maxed out
        prod_2_target = dem_1 + dem_2 - prod_1
        prod_2_target[prod_2_target < 0] = 0
        np.testing.assert_allclose(prod_2, prod_2_target, atol=1e-3, rtol=1e-6)
        np.testing.assert_allclose(prod_1, prod_1_bound)

        # check demand 1 is matching the profile (in check_all_mc_simulations) and that demand 2
        # is maxed out
        checks_all_mc_simulations(solution, results)
        np.testing.assert_allclose(dem_2, dem_2_bound[1], atol=1e-3, rtol=1e-6)

    def test_multi_commodity_simulator_emerge(self):
        import models.emerge.src.example as example

        base_folder = Path(example.__file__).resolve().parent.parent

        solution = run_optimization_problem(
            MultiCommoditySimulatorNoLosses,
            base_folder=base_folder,
            esdl_file_name="emerge_priorities_withoutstorage.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_short.csv",
        )

        bounds = solution.bounds()
        results = solution.extract_results()

        # all_mc_checks cannot be used because max electricity production of windfarm is larger
        # than the max electricity take off at electrolyzer (due to mas gas demand) and the max
        # electricity demand.
        # checks_all_mc_simulations(solution, results)
        tol = 1.0e-6

        esdl_electrolyzer = solution._esdl_assets[
            solution.esdl_asset_name_to_id_map["Electrolyzer_6327"]
        ]
        demand_gas = results["GasDemand_4146.Gas_demand_mass_flow"]
        demand_el = results["ElectricityDemand_f833.Electricity_demand"]
        electrolyzer_power = results["Electrolyzer_6327.Power_consumed"]
        electrolyzer_gas = results["Electrolyzer_6327.Gas_mass_flow_out"]
        windfarm_power = results["WindPark_9074.Electricity_source"]
        windfarm_target = solution.get_timeseries(
            "WindPark_9074.maximum_electricity_source"
        ).values[: len(windfarm_power)]
        # cap on el consumption by electricity_demand (due to cap, 1.3GW) and by electrolyzer
        # (due to cap of gas demand)
        cap_el_consumption = 2.1e9
        windfarm_target = np.minimum(
            np.ones(len(windfarm_target)) * cap_el_consumption, windfarm_target
        )
        np.testing.assert_allclose(windfarm_target, windfarm_power, atol=1.0e-3, rtol=tol)

        check_electrolyzer_efficiency(tol, electrolyzer_gas, electrolyzer_power, esdl_electrolyzer)

        # demand gas maximised when sufficient power available to convert electricity to gas
        cap_electrolyzer_power = (
            cap_el_consumption - bounds["ElectricityDemand_f833.Electricity_demand"][1]
        )
        index_gas_max = windfarm_power >= cap_electrolyzer_power
        np.testing.assert_allclose(
            demand_gas[index_gas_max], bounds["Electrolyzer_6327.Gas_mass_flow_out"][1]
        )
        # due to priority settings, electricity demand only consuming if demand_gas maximised and
        # enough windfarm power
        demand_el_calc = windfarm_power - electrolyzer_power
        demand_el_calc[demand_el_calc < 0] = 0
        np.testing.assert_allclose(demand_el_calc, demand_el, atol=1.0e-3)

    def test_multi_commodity_simulator_emerge_lowprod(self):
        import models.emerge.src.example as example

        base_folder = Path(example.__file__).resolve().parent.parent

        class MCSimulatorShortSmallProd(MultiCommoditySimulatorNoLosses):

            def read(self, variable=None):
                super().read()

                for asset in self.energy_system_components["wind_park"]:
                    new_timeseries = (
                        self.get_timeseries(f"{asset}.maximum_electricity_source").values * 0.5
                    )
                    self.set_timeseries(f"{asset}.maximum_electricity_source", new_timeseries)

            def energy_system_options(self):
                options = super().energy_system_options()

                options["electrolyzer_efficiency"] = (
                    ElectrolyzerOption.LINEARIZED_THREE_LINES_WEAK_INEQUALITY
                )

                return options

        solution = run_optimization_problem(
            MCSimulatorShortSmallProd,
            base_folder=base_folder,
            esdl_file_name="emerge_priorities_withoutstorage_2prod.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_short.csv",
        )

        results = solution.extract_results()
        bounds = solution.bounds()

        tol = 1.0e-6

        checks_all_mc_simulations(solution, results)

        esdl_electrolyzer = solution._esdl_assets[
            solution.esdl_asset_name_to_id_map["Electrolyzer_6327"]
        ]
        demand_gas = results["GasDemand_4146.Gas_demand_mass_flow"]
        demand_el = results["ElectricityDemand_f833.Electricity_demand"]
        electrolyzer_power = results["Electrolyzer_6327.Power_consumed"]
        electrolyzer_gas = results["Electrolyzer_6327.Gas_mass_flow_out"]
        windfarm_power = results["WindPark_9074.Electricity_source"]
        prod_power = results["ElectricityProducer_4850.Electricity_source"]
        prod_power_cap = 3e8  # W

        check_electrolyzer_efficiency(tol, electrolyzer_gas, electrolyzer_power, esdl_electrolyzer)

        # demand gas maximised when sufficient power available to convert electricity to gas
        cap_electrolyzer_power = 444880000.0
        index_gas_max = windfarm_power >= cap_electrolyzer_power
        np.testing.assert_allclose(
            demand_gas[index_gas_max], bounds["Electrolyzer_6327.Gas_mass_flow_out"][1]
        )
        # due to priority settings, electricity demand only consuming if demand_gas maximised and
        # enough windfarm power
        demand_el_calc = windfarm_power - electrolyzer_power
        demand_el_calc[demand_el_calc < 0] = 0
        np.testing.assert_allclose(demand_el_calc, demand_el, atol=1e-5 * max(demand_el))
        # electricity_producer only producing upto its own cap if windfarm_power is insufficient to
        # fullfill demand_gas
        el_prod_calc = np.maximum(
            np.minimum(max(electrolyzer_power) - windfarm_power, prod_power_cap), 0.0
        )
        np.testing.assert_allclose(el_prod_calc, prod_power, atol=1e-5 * max(demand_el))

    def test_multi_commodity_simulator_emerge_storage(self):
        """
        Test to run the multicommodity simulator including a gas_storage.
        Checks:
        - feasibility
        - power conservation
        - efficiency of electrolyzer
        - gas mass balance
        - gas storage charges when additional production of hydrogen is possible and discharges
        when no production is possible.
        """

        import models.emerge.src.example as example

        base_folder = Path(example.__file__).resolve().parent.parent

        solution = run_optimization_problem(
            MultiCommoditySimulatorNoLosses,
            base_folder=base_folder,
            esdl_file_name="emerge_priorities.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_short.csv",
        )

        results = solution.extract_results()

        feasibility_test(solution)
        electric_power_conservation_test(solution, results)

        # checks_all_mc_simulations(solution, results)
        tol = 1.0e-6

        esdl_electrolyzer = solution._esdl_assets[
            solution.esdl_asset_name_to_id_map["Electrolyzer_6327"]
        ]
        demand_gas = results["GasDemand_4146.Gas_demand_mass_flow"]
        electrolyzer_power = results["Electrolyzer_6327.Power_consumed"]
        electrolyzer_gas = results["Electrolyzer_6327.Gas_mass_flow_out"]
        storage_gas_mass_flow = results["GasStorage_9172.Gas_tank_flow"]

        check_electrolyzer_efficiency(tol, electrolyzer_gas, electrolyzer_power, esdl_electrolyzer)

        # gas mass balance including storage
        np.testing.assert_allclose(
            electrolyzer_gas - storage_gas_mass_flow, demand_gas, rtol=tol, atol=tol
        )
        # timestep 7 & 8 electrolyzer should be producing more hydrogen then the demand can take.
        # gas_demand is limited by pipe size.
        # storage is not charging if gas demand is not maximised at bound
        # storage is charging if electrolyzer produces more gas than max of demand.
        gas_demand_bound = solution.bounds()["GasDemand_4146.Gas_demand_mass_flow"]
        storage_not_charging = demand_gas != gas_demand_bound[1]
        np.testing.assert_array_less(storage_gas_mass_flow[storage_not_charging], 0.0 + tol)
        storage_charging = electrolyzer_gas >= gas_demand_bound[1]
        np.testing.assert_array_less(0.0 - tol, storage_gas_mass_flow[storage_charging])

    def test_multi_commodity_simulator_emerge_battery(self):
        """
        Test to run the multicommodity simulator including a battery and gas storage.
        Checks:
        - matching of profiles consumer and producers (checks_all_mc_simulations)
        - feasibility
        - power conservation
        - efficiency of electrolyzer
        - gas mass balance
        - battery charges when over production of electricity and discharges when not enough
        electrical power is produced for the electrolyzer.
        - gas storage is not charged/discharged due to priority level.
        """
        import models.emerge.src.example as example

        base_folder = Path(example.__file__).resolve().parent.parent

        multicommoditysimulatornolossesscaling, logger, logs_list = create_problem_with_debug_info(
            MultiCommoditySimulatorNoLosses
        )

        solution = run_optimization_problem(
            multicommoditysimulatornolossesscaling,
            base_folder=base_folder,
            esdl_file_name="emerge_battery_priorities.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_short.csv",
        )
        problem_scaling_check(logs_list, logger, order_diff=1e7)
        results = solution.extract_results()

        feasibility_test(solution)
        electric_power_conservation_test(solution, results)

        checks_all_mc_simulations(solution, results)
        tol = 1.0e-6

        esdl_electrolyzer = solution._esdl_assets[
            solution.esdl_asset_name_to_id_map["Electrolyzer_6327"]
        ]
        demand_gas = results["GasDemand_4146.Gas_demand_mass_flow"]
        electrolyzer_power = results["Electrolyzer_6327.Power_consumed"]
        electrolyzer_power_bound = solution.bounds()["Electrolyzer_6327.Power_consumed"][1]
        electrolyzer_gas = results["Electrolyzer_6327.Gas_mass_flow_out"]
        windfarm_power = results["WindPark_9074.Electricity_source"]
        storage_gas_mass_flow = results["GasStorage_9172.Gas_tank_flow"]
        battery_power = results["Battery_4688.ElectricityIn.Power"]

        check_electrolyzer_efficiency(tol, electrolyzer_gas, electrolyzer_power, esdl_electrolyzer)

        # gas mass balance including storage
        np.testing.assert_allclose(
            electrolyzer_gas - storage_gas_mass_flow, demand_gas, rtol=tol, atol=tol
        )

        # check battery only discharged if not enough windpower for electrolyzer
        discharge_battery = windfarm_power < electrolyzer_power_bound
        np.testing.assert_array_less(battery_power[discharge_battery], 0.0 + tol)
        charge_battery = windfarm_power > electrolyzer_power_bound
        np.testing.assert_array_less(0.0, battery_power[charge_battery])

        # gas_storage not used due to low priority
        np.testing.assert_allclose(storage_gas_mass_flow[1:], 0.0, atol=tol)

    def test_multi_commodity_simulator_emerge_head_losses(self):
        """
        Test to run the multicommodity simulator including a battery and gas storage.
        Checks:
        - feasibility
        - power conservation
        - efficiency of electrolyzer
        - gas mass balance
        - battery charges when over production of electricity and discharges when not enough
        electrical power is produced for the electrolyzer.
        - gas storage is not charged/discharged due to priority level.
        """
        import models.emerge.src.example as example

        base_folder = Path(example.__file__).resolve().parent.parent

        solution = run_optimization_problem(
            MultiCommoditySimulator,
            base_folder=base_folder,
            esdl_file_name="emerge_battery_priorities.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_short_2.csv",
        )

        results = solution.extract_results()
        parameters = solution.parameters(0)

        feasibility_test(solution)
        electric_power_conservation_test(solution, results)

        checks_all_mc_simulations(solution, results)
        tol = 1.0e-6

        esdl_electrolyzer = solution._esdl_assets[
            solution.esdl_asset_name_to_id_map["Electrolyzer_6327"]
        ]
        demand_gas = results["GasDemand_4146.Gas_demand_mass_flow"]
        electrolyzer_power = results["Electrolyzer_6327.Power_consumed"]
        electrolyzer_power_bound = solution.bounds()["Electrolyzer_6327.Power_consumed"][1]
        electrolyzer_gas = results["Electrolyzer_6327.Gas_mass_flow_out"]
        windfarm_power = results["WindPark_9074.Electricity_source"]
        storage_gas_mass_flow = results["GasStorage_9172.Gas_tank_flow"]
        battery_power = results["Battery_4688.ElectricityIn.Power"]

        check_electrolyzer_efficiency(tol, electrolyzer_gas, electrolyzer_power, esdl_electrolyzer)

        # gas mass balance including storage
        np.testing.assert_allclose(
            electrolyzer_gas - storage_gas_mass_flow, demand_gas, rtol=tol, atol=tol
        )

        # check battery only discharged if not enough windpower for electrolyzer
        discharge_battery = windfarm_power < electrolyzer_power_bound
        np.testing.assert_array_less(battery_power[discharge_battery], 1.0)
        charge_battery = windfarm_power > electrolyzer_power_bound
        np.testing.assert_array_less(0.0, battery_power[charge_battery])

        # gas_storage not used due to low priority
        np.testing.assert_allclose(storage_gas_mass_flow[1:], 0.0, atol=tol)

        # linearized dw_headloss calculations
        linear_lines = 5
        v_max = solution.gas_network_settings["maximum_velocity"]
        velocities = np.linspace(0, v_max, linear_lines + 1)

        for pipe in solution.energy_system_components.get("gas_pipe"):
            length = parameters[f"{pipe}.length"]
            diameter = parameters[f"{pipe}.diameter"]
            vol_flow_rate = results[f"{pipe}.GasIn.Q"]
            v_pipe = vol_flow_rate / (3.14 * (diameter / 2) ** 2)
            wall_roughness = solution.energy_system_options()["wall_roughness"]
            head_loss = results[f"{pipe}.dH"]
            head_loss_full_var = results[f"{pipe}.__head_loss"]
            # If this test fails there is most likely a scaling issue.
            indexes = np.abs(v_pipe) > 1e-11
            indexes[0] = False
            np.testing.assert_allclose(
                np.abs(np.asarray(head_loss[indexes])), head_loss_full_var[indexes]
            )
            for i in range(1, len(v_pipe)):
                v = v_pipe[i]
                line_num = velocities.searchsorted(abs(v))
                if abs(v) > 1e-8:
                    dw_headloss_max = darcy_weisbach.head_loss(
                        velocities[line_num],
                        diameter,
                        length,
                        wall_roughness,
                        20,
                        NetworkSettings.NETWORK_TYPE_HYDROGEN,
                        15e5,
                    )
                    dw_headloss_min = darcy_weisbach.head_loss(
                        velocities[line_num - 1],
                        diameter,
                        length,
                        wall_roughness,
                        20,
                        NetworkSettings.NETWORK_TYPE_HYDROGEN,
                        15e5,
                    )
                    a = (dw_headloss_max - dw_headloss_min) / (
                        velocities[line_num] - velocities[line_num - 1]
                    )
                    b = dw_headloss_min - a * velocities[line_num - 1]
                    headloss_calc = a * v + b
                    np.testing.assert_allclose(abs(head_loss[i]), headloss_calc, 0.1)

    def test_multi_commodity_simulator_sequential_staged(self):
        """
        Test to run the multicommodity simulator including a battery and gas storage using
        the sequential staged optimization approach. The results between the staged and unstaaged
        approach should be equal if the assets in the system do not have an internal state that
        makes timesteps interdependent (like Stored_energy for a battery) or when the bounds on
        those states do not limit the problem.

        Checks:
        - that the staged approach results in same values as the unstaged approach.
        - Verify that under limited case bounds are set correctly.
        """
        import models.emerge.src.example as example

        base_folder = Path(example.__file__).resolve().parent.parent

        solution_staged_unbounded = run_sequatially_staged_simulation(
            multi_commodity_simulator_class=MultiCommoditySimulatorNoLosses,
            simulation_window_size=20,
            base_folder=base_folder,
            esdl_file_name="emerge_battery_priorities.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_short.csv",
        )

        results_staged = solution_staged_unbounded.results

        solution_unstaged = run_optimization_problem(
            MultiCommoditySimulatorNoLosses,
            base_folder=base_folder,
            esdl_file_name="emerge_battery_priorities.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_short.csv",
        )

        results_unstaged = solution_unstaged.extract_results()

        # Checking that the results are the same.
        for key, value in results_unstaged.items():
            value_staged = results_staged[key]
            if len(key.split("__")) > 1 and key.split("__")[1] == "gas_flow_direct_var":
                # For the scenario when Q is -0 and 0 then gas_flow_direct_var 0 or 1 for the same
                # volumetric flow rate of zero. So only check gas_flow_direct_var when Q != zero
                zero_staged = results_staged[f"{key.split('__')[0]}.GasIn.Q"] != 0
                zero_unstaged = results_unstaged[f"{key.split('__')[0]}.GasIn.Q"] != 0
                np.testing.assert_allclose(value[zero_staged], value_staged[zero_unstaged])
            else:
                np.testing.assert_allclose(value, value_staged, atol=1e-4)

        solution_staged_bounded = run_sequatially_staged_simulation(
            multi_commodity_simulator_class=MultiCommoditySimulatorNoLosses,
            simulation_window_size=20,
            base_folder=base_folder,
            esdl_file_name="emerge_battery_priorities_limited_capacity.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_short.csv",
        )

        results_staged_bounded = solution_staged_bounded.results

        solution_unstaged_bounded = run_optimization_problem(
            MultiCommoditySimulatorNoLosses,
            base_folder=base_folder,
            esdl_file_name="emerge_battery_priorities_limited_capacity.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_short.csv",
        )

        results_unstaged_bounded = solution_unstaged_bounded.extract_results()

        check_different = np.sum(
            np.abs(
                results_staged_bounded["Battery_4688.Stored_electricity"]
                - results_unstaged_bounded["Battery_4688.Stored_electricity"]
            )
        )

        assert check_different > 0

        class MultiCommoditySimulatorNoLossesWin(MultiCommoditySimulatorNoLosses):
            def times(self, variable=None) -> np.ndarray:
                return super().times(variable)[:20]

        solution_unstaged_bounded_win = run_optimization_problem(
            MultiCommoditySimulatorNoLossesWin,
            base_folder=base_folder,
            esdl_file_name="emerge_battery_priorities_limited_capacity.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_short.csv",
        )

        results_unstaged_bounded_win = solution_unstaged_bounded_win.extract_results()

        np.testing.assert_allclose(
            results_staged_bounded["Battery_4688.Stored_electricity"][19],
            results_unstaged_bounded_win["Battery_4688.Stored_electricity"][-1],
        )


if __name__ == "__main__":
    import time

    start_time = time.time()
    a = TestMultiCommoditySimulator()
    # a.test_multi_commodity_simulator_priorities_el()
    # a.test_multi_commodity_simulator_prod_profile()
    # a.test_multi_commodity_simulator_emerge()
    # a.test_multi_commodity_simulator_emerge_lowprod()
    # a.test_multi_commodity_simulator_emerge_head_losses()
    a.test_multi_commodity_simulator_sequential_staged()

    print("Execution time: " + time.strftime("%M:%S", time.gmtime(time.time() - start_time)))



from unittest import TestCase
from unittest.mock import MagicMock

from mesido.workflows.grow_workflow import estimate_and_update_progress_status


class TestProgress(TestCase):

    # 1
    def estimate_progress_nhl_ts1_s1_p1(self, task_quantity_perc_completed: float, msg: str):
        self.assertEqual(
            [task_quantity_perc_completed, msg],
            [1.0 / 2.0, "Optimization task 1.0 out of 2.0 has completed"],
        )

    # 2
    def estimate_progress_nhl_ts1_s1_p2(self, task_quantity_perc_completed: float, msg: str):
        self.assertEqual(
            [task_quantity_perc_completed, msg],
            [2.0 / 2.0, "Optimization task 2.0 out of 2.0 has completed"],
        )
        return None

    # 3
    def estimate_progress_hl_ts2_s1_phl(self, task_quantity_perc_completed: float, msg: str):
        return None

    # 4.1
    def estimate_progress_nhl_ts2_s1_p1(self, task_quantity_perc_completed: float, msg: str):
        self.assertEqual(
            [task_quantity_perc_completed, msg],
            [1.0 / 4.0, "Optimization task 1.0 out of 4.0 has completed"],
        )
        return None

    # 4.2
    def estimate_progress_nhl_ts2_s1_p2(self, task_quantity_perc_completed: float, msg: str):
        self.assertEqual(
            [task_quantity_perc_completed, msg],
            [2.0 / 4.0, "Optimization task 2.0 out of 4.0 has completed"],
        )
        return None

    # 5.1
    def estimate_progress_nhl_ts2_s2_p1(self, task_quantity_perc_completed: float, msg: str):
        self.assertEqual(
            [task_quantity_perc_completed, msg],
            [3.0 / 4.0, "Optimization task 3.0 out of 4.0 has completed"],
        )
        return None

    # 5.2
    def estimate_progress_nhl_ts2_s2_p2(self, task_quantity_perc_completed: float, msg: str):
        self.assertEqual(
            [task_quantity_perc_completed, msg],
            [4.0 / 4.0, "Optimization task 4.0 out of 4.0 has completed"],
        )
        return None

    # 6
    def estimate_progress_hl_ts2_s2_phl(self, task_quantity_perc_completed: float, msg: str):
        self.assertEqual(
            [task_quantity_perc_completed, msg],
            [5.0 / 6.0, "Optimization task 5.0 out of 6.0 has completed"],
        )
        return None

    # 7
    def estimate_progress_hl_ts2_s2_php(self, task_quantity_perc_completed: float, msg: str):
        self.assertEqual(
            [task_quantity_perc_completed, msg],
            [6.0 / 6.0, "Optimization task 6.0 out of 6.0 has completed"],
        )
        return None


class TestEstimateAndUpdateProgressStatus(TestCase):

    def test_progess(self):
        """
        Test the progress status of the grow workflow for the different scenarios are calculated
        correctly.

        Checks:
        - #1: Priority = 1, total stages = 1, stage = 1, minimize_head_losses = False -> macth heat
        demand
        - #2: Priority = 2, total stages = 1, stage = 1, minimize_head_losses = False -> minimize
        TCO

        - #3: Priority = 2**31 - 2, total stages = 2, stage = 1, minimize_head_losses = True -> give
        error since the function does not cater for this scenario

        - #4.1: Priority = 1, total stages = 2, stage = 1 -> match heating demand
        - #4.2: Priority = 1, total stages = 2, stage 1 = -> minimize TCO
        - #5.1: Priority = 1, total stages = 2, stage = 2 -> match heating demand
        - #5.2: Priority = 2, total stages = 2, stage = 2 -> minimize TCO

        - #6 Priority = 2**31 - 2, total stages = 2, stage = 2, minimize_head_losses = True -> head
        loss optimization
        - #7 Priority = 2**31 - 1, total stages = 2, stage = 2, minimize_head_losses = True ->
        hydraulic power optimization


        """

        mock_endscenario_class = MagicMock()
        test_progresss = TestProgress()

        # 1
        priority = 1
        mock_endscenario_class._total_stages = 1
        mock_endscenario_class._stage = 1
        mock_endscenario_class.heat_network_settings = {"minimize_head_losses": False}
        mock_endscenario_class._workflow_progress_status = (
            test_progresss.estimate_progress_nhl_ts1_s1_p1
        )
        estimate_and_update_progress_status(mock_endscenario_class, priority)

        # 2
        priority = 2
        mock_endscenario_class._total_stages = 1
        mock_endscenario_class._stage = 1
        mock_endscenario_class.heat_network_settings = {"minimize_head_losses": False}
        mock_endscenario_class._workflow_progress_status = (
            test_progresss.estimate_progress_nhl_ts1_s1_p2
        )
        estimate_and_update_progress_status(mock_endscenario_class, priority)

        # 3
        priority = 2**31 - 2
        mock_endscenario_class._total_stages = 2
        mock_endscenario_class._stage = 1
        mock_endscenario_class.heat_network_settings = {"minimize_head_losses": True}
        mock_endscenario_class._workflow_progress_status = (
            test_progresss.estimate_progress_hl_ts2_s1_phl
        )
        with self.assertRaises(SystemExit) as cm:
            estimate_and_update_progress_status(mock_endscenario_class, priority)
        self.assertEqual(
            cm.exception.args[0],
            "The function does not cater for stage number:1 & priority:2147483646",
        )

        # 4.1
        priority = 1
        mock_endscenario_class._total_stages = 2
        mock_endscenario_class._stage = 1
        mock_endscenario_class.heat_network_settings = {"minimize_head_losses": False}
        mock_endscenario_class._workflow_progress_status = (
            test_progresss.estimate_progress_nhl_ts2_s1_p1
        )
        estimate_and_update_progress_status(mock_endscenario_class, priority)

        # 4.2
        priority = 2
        mock_endscenario_class._total_stages = 2
        mock_endscenario_class._stage = 1
        mock_endscenario_class.heat_network_settings = {"minimize_head_losses": False}
        mock_endscenario_class._workflow_progress_status = (
            test_progresss.estimate_progress_nhl_ts2_s1_p2
        )
        estimate_and_update_progress_status(mock_endscenario_class, priority)

        # 5.1
        priority = 1
        mock_endscenario_class._total_stages = 2
        mock_endscenario_class._stage = 2
        mock_endscenario_class.heat_network_settings = {"minimize_head_losses": False}
        mock_endscenario_class._workflow_progress_status = (
            test_progresss.estimate_progress_nhl_ts2_s2_p1
        )
        estimate_and_update_progress_status(mock_endscenario_class, priority)

        # 5.2
        priority = 2
        mock_endscenario_class._total_stages = 2
        mock_endscenario_class._stage = 2
        mock_endscenario_class.heat_network_settings = {"minimize_head_losses": False}
        mock_endscenario_class._workflow_progress_status = (
            test_progresss.estimate_progress_nhl_ts2_s2_p2
        )
        estimate_and_update_progress_status(mock_endscenario_class, priority)

        # 6
        priority = 2**31 - 2
        mock_endscenario_class._total_stages = 2
        mock_endscenario_class._stage = 2
        mock_endscenario_class.heat_network_settings = {"minimize_head_losses": True}
        mock_endscenario_class._workflow_progress_status = (
            test_progresss.estimate_progress_hl_ts2_s2_phl
        )
        estimate_and_update_progress_status(mock_endscenario_class, priority)

        # 7
        priority = 2**31 - 1
        mock_endscenario_class._total_stages = 2
        mock_endscenario_class._stage = 2
        mock_endscenario_class.heat_network_settings = {"minimize_head_losses": True}
        mock_endscenario_class._workflow_progress_status = (
            test_progresss.estimate_progress_hl_ts2_s2_php
        )
        estimate_and_update_progress_status(mock_endscenario_class, priority)


if __name__ == "__main__":
    a = TestEstimateAndUpdateProgressStatus()
    a.test_progess()



from mesido.pycml.pycml_mixin import add_variables_documentation_automatically

from ._non_storage_component import _NonStorageComponent


@add_variables_documentation_automatically
class ControlValve(_NonStorageComponent):
    """
    The control valve is a component to create pressure drop. We allow the control valve to create
    pressure drop for flow in both directions. Note that we set the absolute head loss symbol in
    the HeatMixin.

    Variables created:
        {add_variable_names_for_documentation_here}

    Parameters:
        name : The name of the asset. \n
        modifiers : Dictionary with asset information.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "control_valve"

        self.add_equation(
            (self.HeatOut.Hydraulic_power - self.HeatIn.Hydraulic_power)
            / (self.Q_nominal * self.nominal_pressure)
        )

        self.add_equation((self.HeatIn.Heat - self.HeatOut.Heat) / self.Heat_nominal)



from mesido.pycml import Connector, Variable
from mesido.pycml.component_library.milp._internal import HeatComponent
from mesido.pycml.pycml_mixin import add_variables_documentation_automatically


@add_variables_documentation_automatically
class HeatPort(HeatComponent, Connector):
    """
    The HeatPort is used to model the variables at an in or outgoing port of a component. For the
    HeatMixin we model thermal Power (Heat [W]), flow (Q [m3/s]) and head (H [m]) at every port in
    the network.

    Variables created:
        {add_variable_names_for_documentation_here}

    Parameters:
        name : The name of the asset. \n
        modifiers : Dictionary with asset information.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.add_variable(Variable, "Heat")
        self.add_variable(Variable, "Q")
        self.add_variable(Variable, "H")
        self.add_variable(Variable, "Hydraulic_power")



from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.util import run_esdl_mesido_optimization

import numpy as np

from utils_tests import (
    demand_matching_test,
    electric_power_conservation_test,
    energy_conservation_test,
    heat_to_discharge_test,
)


class TestElecBoiler(TestCase):

    def asset_cost_calculation_tests(self, solution, results):
        # Check the cost components
        for asset in [
            *solution.energy_system_components.get("heat_source", []),
        ]:
            esdl_asset = solution.esdl_assets[solution.esdl_asset_name_to_id_map[f"{asset}"]]
            costs_esdl_asset = esdl_asset.attributes["costInformation"]

            # investment cost
            investment_cost_info = costs_esdl_asset.investmentCosts.value
            investment_cost = investment_cost_info * results[f"{asset}__max_size"] / 1.0e6
            np.testing.assert_allclose(
                investment_cost, results[f"{asset}__investment_cost"], atol=1.0e-7
            )

            # installation cost
            installation_cost = costs_esdl_asset.installationCosts.value
            np.testing.assert_allclose(installation_cost, results[f"{asset}__installation_cost"])

            # variable operational cost
            timesteps_hr = np.diff(solution.times()) / 3600.0
            variable_operational_cost = 0.0
            var_op_costs = costs_esdl_asset.variableOperationalCosts.value
            for ii in range(1, len(solution.times())):
                variable_operational_cost += (
                    var_op_costs
                    * results[f"{asset}.Power_consumed"][ii]
                    * timesteps_hr[ii - 1]
                    / 1e6
                )
            np.testing.assert_allclose(
                variable_operational_cost, results[f"{asset}__variable_operational_cost"]
            )

    def test_elec_heat_source_elec(self):
        """
        This tests checks the elec heat sourc elec for the standard checks and the energy
        conservation over the commodity change.

        Checks:
        1. demand is matched
        2. energy conservation in the network
        3. heat to discharge
        4. energy conservation over the heat and electricity commodity
        5. ElectricBoiler cost components
        """
        import models.source_pipe_sink.src.double_pipe_heat as example
        from models.source_pipe_sink.src.double_pipe_heat import SourcePipeSink

        base_folder = Path(example.__file__).resolve().parent.parent

        heat_problem = run_esdl_mesido_optimization(
            SourcePipeSink,
            base_folder=base_folder,
            esdl_file_name="sourcesink_with_eboiler.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.csv",
        )
        results = heat_problem.extract_results()
        parameters = heat_problem.parameters(0)

        demand_matching_test(heat_problem, results)
        energy_conservation_test(heat_problem, results)
        heat_to_discharge_test(heat_problem, results)
        electric_power_conservation_test(heat_problem, results)

        np.testing.assert_array_less(0.0, results["ElectricBoiler_9aab.Heat_source"])
        np.testing.assert_array_less(0.0, results["ElectricityProducer_4dde.ElectricityOut.Power"])
        np.testing.assert_allclose(
            parameters["ElectricBoiler_9aab.efficiency"]
            * results["ElectricBoiler_9aab.Power_consumed"],
            results["ElectricBoiler_9aab.Heat_source"],
        )

        # Check the cost components of ElectricBoiler
        self.asset_cost_calculation_tests(heat_problem, results)

    def test_heat_source_elec(self):
        """
        This tests checks the heat sourc elec for the standard checks and the energy conservation
        over the commodity change.

        Checks:
        1. demand is matched
        2. energy conservation in the network
        3. heat to discharge
        4. energy conservation over the heat and electricity commodity
        5. ElectricBoiler cost components
        """
        import models.source_pipe_sink.src.double_pipe_heat as example
        from models.source_pipe_sink.src.double_pipe_heat import SourcePipeSink

        base_folder = Path(example.__file__).resolve().parent.parent

        heat_problem = run_esdl_mesido_optimization(
            SourcePipeSink,
            base_folder=base_folder,
            esdl_file_name="sourcesink_with_eboiler_no_elec.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.csv",
        )
        results = heat_problem.extract_results()
        parameters = heat_problem.parameters(0)

        demand_matching_test(heat_problem, results)
        energy_conservation_test(heat_problem, results)
        heat_to_discharge_test(heat_problem, results)

        np.testing.assert_array_less(0.0, results["ElectricBoiler_9aab.Heat_source"])
        np.testing.assert_allclose(
            parameters["ElectricBoiler_9aab.efficiency"]
            * results["ElectricBoiler_9aab.Power_consumed"],
            results["ElectricBoiler_9aab.Heat_source"],
        )

        # Check the cost components of ElectricBoiler
        self.asset_cost_calculation_tests(heat_problem, results)

    def test_air_water_hp_elec(self):
        """
        This tests checks the air-water hp elec for the standard checks and the energy conservation
        over the commodity change.

        Checks:
        1. demand is matched
        2. energy conservation in the network
        3. heat to discharge
        4. energy conservation over the commodity
        """
        import models.source_pipe_sink.src.double_pipe_heat as example
        from models.source_pipe_sink.src.double_pipe_heat import SourcePipeSink

        base_folder = Path(example.__file__).resolve().parent.parent

        heat_problem = run_esdl_mesido_optimization(
            SourcePipeSink,
            base_folder=base_folder,
            esdl_file_name="sourcesink_withHP.esdl",
            esdl_parser=ESDLFileParser,
            profile_reader=ProfileReaderFromFile,
            input_timeseries_file="timeseries_import.csv",
        )
        results = heat_problem.extract_results()
        parameters = heat_problem.parameters(0)

        demand_matching_test(heat_problem, results)
        energy_conservation_test(heat_problem, results)
        heat_to_discharge_test(heat_problem, results)
        electric_power_conservation_test(heat_problem, results)

        np.testing.assert_array_less(0.0, results["HeatPump_d8fd.Heat_source"])
        np.testing.assert_array_less(0.0, results["ElectricityProducer_4dde.ElectricityOut.Power"])
        np.testing.assert_array_less(
            parameters["HeatPump_d8fd.cop"] * results["HeatPump_d8fd.Power_elec"],
            results["HeatPump_d8fd.Heat_source"] + 1.0e-6,
        )

        # Check how variable operation cost is calculated
        np.testing.assert_allclose(
            parameters["HeatPump_d8fd.variable_operational_cost_coefficient"]
            * sum(results["HeatPump_d8fd.Heat_source"][1:])
            / parameters["HeatPump_d8fd.cop"],
            results["HeatPump_d8fd__variable_operational_cost"],
        )


if __name__ == "__main__":
    TestElecBoiler = TestElecBoiler()
    TestElecBoiler.test_elec_heat_source_elec()
    TestElecBoiler.test_heat_source_elec()
    TestElecBoiler.test_air_water_hp_elec()



import math

from mesido.pycml import Variable

from numpy import nan

from ._fluid_properties_component import _FluidPropertiesComponent


class Buffer(_FluidPropertiesComponent):
    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "heat_buffer"

        self.Q_nominal = 1.0
        self.head_loss = 0.0

        # Source for U estimates:
        # https://www.spiraxsarco.com/learn-about-steam/steam-engineering-principles-and-heat-transfer/energy-consumption-of-tanks-and-vats
        self.heat_transfer_coeff = 1.0
        self.height = 5.0
        self.radius = 10.0
        self.volume = math.pi * self.radius**2 * self.height
        # The hot/cold tank can have a lower bound on its volume.
        # Meaning that they might always be, for e.g., 5% full.
        self.min_fraction_tank_volume = 0.05

        # if the tank is plced on ground, ignore surface of bottom of tank
        self.surface = 2 * math.pi * self.radius**2 + 2 * math.pi * self.radius * self.height
        self.T_outside = 10.0

        # Nominals
        self.nom_tank_volume = self.volume / 2
        self.nom_heat = self.cp * self.rho * self.dT

        # Initial values
        self.init_V_hot_tank = nan
        self.init_T_hot_tank = self.T_supply
        self.init_T_cold_tank = self.T_return

        # Volume of the tanks
        self.add_variable(
            Variable,
            "V_hot_tank",
            min=self.volume * self.min_fraction_tank_volume,
            max=self.volume,
            nominal=self.nom_tank_volume,
        )
        self.add_variable(
            Variable,
            "V_cold_tank",
            min=self.volume * self.min_fraction_tank_volume,
            max=self.volume,
            nominal=self.nom_tank_volume,
        )

        # Temperature in the tanks
        self.add_variable(Variable, "T_hot_tank", min=self.T_outside, nominal=self.T_supply)
        self.add_variable(Variable, "T_cold_tank", min=self.T_outside, nominal=self.T_return)

        # Alias variables for the in/out pipes.
        self.add_variable(Variable, "Q_hot_pipe", nominal=self.Q_nominal)
        self.add_variable(Variable, "Q_cold_pipe", nominal=self.Q_nominal)
        self.add_variable(Variable, "T_hot_pipe", nominal=self.T_supply)
        self.add_variable(Variable, "T_cold_pipe", nominal=self.T_return)

        # Buffer is modelled with an hot and a cold tank.
        # The hot tank is connected to the supply line, while the cold one to the return line.
        # The total volume of the system is constant.
        # Assumption: constant density and milp capacity
        # Q_hot_pipe is positive when the buffer is charging, negative if discharching.
        # Note that, as volume is constant, the amount of hot and cold water discharged
        # must be equal.
        # For convention, we assume that Q_hot_pipe and Q_cold_pipe have the same sign, i.e.,
        # Q_hot_pipe = Q_cold_pipe.

        # We couple the flows with the volume in the tanks
        self.add_equation((self.der(self.V_hot_tank) - self.Q_hot_pipe))
        self.add_equation((self.der(self.V_cold_tank) + self.Q_cold_pipe))

        # The following relationships are set in the mixin:

        # * Temperatures of the in/out pipes which are direction dependend.
        # If Q_hot_pipe < 0.0, T_hot_pipe = T_hot_tank.
        # Else, T_cold_pipe = T_cold_tank.
        # T_hot_pipe and T_cold pipe must be the one coming from the network
        # in the other timesteps.

        # * Heat equations:
        # der(V_hot_tank * T_hot_tank) - (Q_hot_pipe * T_hot_pipe) = 0.0;
        # der(V_cold_tank * T_hot_tank) - (Q_cold_pipe * T_cold_pipe) = 0.0;

        # * Flows in/out:
        # recall that Q_hot_pipe and Q_cold_pipe have the same sign.
        # To compensate for pipe orientations, we have:
        # hot_pipe_orientation * QTHIn.Q = Q_hot_pipe;
        # -1 * cold_pipe_orientation * QTHOut.Q = Q_cold_pipe;

        # * Initial conditions at t0:
        # set volume/temperature of the hot and cold tank as bounds.

        # Aliases
        self.add_equation(self.QTHIn.T - self.T_hot_pipe)
        self.add_equation(self.QTHOut.T - self.T_cold_pipe)



from enum import IntEnum


class CheckValveStatus(IntEnum):
    """
    Enumeration for the possible status a check valve can have.
    """

    CLOSED = 0
    OPEN = 1


class ControlValveDirection(IntEnum):
    """
    Enumeration for the possible directions a control valve can have.
    """

    NEGATIVE = -1
    POSITIVE = 1


class PipeFlowDirection(IntEnum):
    """
    Enumeration for the possible directions a pipe can have.
    """

    NEGATIVE = -1
    DISABLED = 0
    POSITIVE = 1


class NodeConnectionDirection(IntEnum):
    """
    Enumeration for the orientation of a pipe connected to a node, or of the
    flow into a node.
    """

    OUT = -1
    IN = 1



from mesido.pycml import Variable
from mesido.pycml.pycml_mixin import add_variables_documentation_automatically

from numpy import nan

from .electricity_base import ElectricityPort
from .._internal import BaseAsset
from .._internal.electricity_component import ElectricityComponent


@add_variables_documentation_automatically
class ElectricityDemand(ElectricityComponent, BaseAsset):
    """
    The electricity demand models consumption of electrical power. We set an equality constriant
    in to enforce the minimum voltage and the associated power at the demand. This allows us to
    overestimate the power losses in the rest of the network.

    Variables created:
        {add_variable_names_for_documentation_here}

    Parameters:
        name : The name of the asset. \n
        modifiers : Dictionary with asset information.
    """

    def __init__(self, name, **modifiers):
        super().__init__(name, **modifiers)

        self.component_type = "electricity_demand"
        self.min_voltage = nan
        self.elec_power_nominal = nan
        self.id_mapping_carrier = -1

        self.add_variable(ElectricityPort, "ElectricityIn")
        self.add_variable(Variable, "Electricity_demand", min=0.0, nominal=self.elec_power_nominal)

        self.add_equation(
            ((self.ElectricityIn.Power - self.Electricity_demand) / self.elec_power_nominal)
        )



from .electricity.electricity_cable import ElectricityCable
from .electricity.electricity_demand import ElectricityDemand
from .electricity.electricity_node import ElectricityNode
from .electricity.electricity_source import ElectricitySource
from .electricity.electricity_storage import ElectricityStorage
from .electricity.heat_pump_elec import HeatPumpElec
from .electricity.solarpv import SolarPV
from .electricity.transformer import Transformer
from .electricity.windpark import WindPark
from .gas.compressor import Compressor
from .gas.gas_demand import GasDemand
from .gas.gas_node import GasNode
from .gas.gas_pipe import GasPipe
from .gas.gas_source import GasSource
from .gas.gas_substation import GasSubstation
from .gas.gas_tank_storage import GasTankStorage
from .heat.air_water_heat_pump import AirWaterHeatPump
from .heat.airco import Airco
from .heat.ates import ATES
from .heat.check_valve import CheckValve
from .heat.cold_demand import ColdDemand
from .heat.control_valve import ControlValve
from .heat.geothermal_source import GeothermalSource
from .heat.heat_buffer import HeatBuffer
from .heat.heat_demand import HeatDemand
from .heat.heat_exchanger import HeatExchanger
from .heat.heat_four_port import HeatFourPort
from .heat.heat_pipe import HeatPipe
from .heat.heat_port import HeatPort
from .heat.heat_pump import HeatPump
from .heat.heat_source import HeatSource
from .heat.heat_source_elec import HeatSourceElec
from .heat.heat_source_gas import HeatSourceGas
from .heat.heat_two_port import HeatTwoPort
from .heat.low_temperature_ates import LowTemperatureATES
from .heat.node import Node
from .heat.pump import Pump
from .multicommodity.airwater_heat_pump_elec import AirWaterHeatPumpElec
from .multicommodity.elec_heat_source_elec import ElecHeatSourceElec
from .multicommodity.electrolyzer import Electrolyzer
from .multicommodity.gas_heat_source_gas import GasHeatSourceGas

__all__ = [
    "Airco",
    "AirWaterHeatPump",
    "AirWaterHeatPumpElec",
    "ATES",
    "HeatBuffer",
    "CheckValve",
    "ColdDemand",
    "Compressor",
    "ControlValve",
    "HeatDemand",
    "HeatSourceElec",
    "ElecHeatSourceElec",
    "ElectricityCable",
    "ElectricityDemand",
    "ElectricityNode",
    "ElectricitySource",
    "ElectricityStorage",
    "Electrolyzer",
    "HeatSourceGas",
    "GasHeatSourceGas",
    "GasDemand",
    "GasNode",
    "GasPipe",
    "GasSource",
    "GasSubstation",
    "GasTankStorage",
    "GeothermalSource",
    "HeatExchanger",
    "HeatFourPort",
    "HeatPipe",
    "HeatPort",
    "HeatPump",
    "HeatPumpElec",
    "HeatTwoPort",
    "LowTemperatureATES",
    "Node",
    "Pump",
    "HeatSource",
    "SolarPV",
    "Transformer",
    "WindPark",
]



import logging
from typing import List

import esdl

from mesido.network_common import NetworkSettings
from mesido.pipe_class import PipeClass

import numpy as np

from rtctools.optimization.collocated_integrated_optimization_problem import (
    CollocatedIntegratedOptimizationProblem,
)

logger = logging.getLogger("mesido")


class ESDLAdditionalVarsMixin(CollocatedIntegratedOptimizationProblem):
    __temperature_options = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def read(self):
        super().read()
        parameters = self.parameters(0)
        bounds = self.bounds()

        # ------------------------------------------------------------------------------------------
        # Limit available pipe classes
        # TODO: cater for varying temperature when limiting pipe classes below
        # Here we do a check between the available pipe classes and the connected assets. For
        # demands this refers to the demand profiles and for producers to their maximum capacity.
        # This is to ensure that we don't have unneeded large amount of available pipe classes
        # for pipes connected to smaller demands and producers.
        # TODO: add the same for electricity ones we have proper support for that in the ESDLMixin

        def __limit_list_available_heat_pipe_classes_max_power(
            max_power: float, demand: bool
        ) -> None:
            """
            Limits the available heat pipe classes for optimization based on the maximum power
            that can be connected.
            """
            new_pcs = []
            found_pc_large_enough = False
            for pc in self.pipe_classes(connected_asset):
                if not found_pc_large_enough:
                    new_pcs.append(pc)
                if (
                    new_pcs[-1].maximum_discharge
                    * parameters[f"{asset}.cp"]
                    * parameters[f"{asset}.rho"]
                    * (parameters[f"{asset}.T_supply"] - parameters[f"{asset}.T_return"])
                ) >= max_power:
                    found_pc_large_enough = True
            self.remove_dn0(new_pcs, is_there_always_mass_flow)
            self._override_pipe_classes[connected_asset] = new_pcs

            if not self.is_hot_pipe(self.hot_to_cold_pipe(connected_asset)):
                if self.has_related_pipe(connected_asset):
                    self._override_pipe_classes[self.hot_to_cold_pipe(connected_asset)] = new_pcs
                else:
                    # TODO: temporarily fix for pipes that don't have related attribute,
                    # but are connected to demands, to still limit the classes. To be
                    # fixed in Topology in separate PR.
                    if demand:
                        port_1, port_2 = "Out", "In"
                    else:
                        port_1, port_2 = "In", "Out"
                    alias = [
                        x
                        for x in self.alias_relation.aliases(f"{asset}.Heat{port_1}.Heat")
                        if not x.startswith(asset) and x.endswith(".Heat")
                    ][0]
                    connected_asset_2 = alias[: -len(f".Heat{port_2}.Heat")]
                    self._override_pipe_classes[connected_asset_2] = new_pcs

        if len(self.temperature_carriers().items()) == 0:
            for asset, (
                connected_asset,
                _orientation,
            ) in self.energy_system_topology.demands.items():
                # TODO: add test case for gas once optional gas pipes are used
                if asset in self.energy_system_components.get("gas_demand", []):
                    try:
                        max_demand_g_s = min(
                            max(self.get_timeseries(f"{asset}.target_gas_demand").values),
                            bounds[f"{asset}.Gas_demand_mass_flow"][1],
                        )
                    except KeyError:
                        max_demand_g_s = bounds[f"{asset}.Gas_demand_mass_flow"][1]
                    is_there_always_mass_flow = False
                    try:
                        if min(self.get_timeseries(f"{asset}.target_gas_demand").values) > 0.0:
                            is_there_always_mass_flow = True
                    except KeyError:
                        is_there_always_mass_flow = False

                    new_pcs = []
                    found_pc_large_enough = False
                    for pc in self.gas_pipe_classes(connected_asset):
                        if not found_pc_large_enough:
                            new_pcs.append(pc)
                            if (
                                new_pcs[-1].maximum_discharge
                                * self.parameters(0)[f"{connected_asset}.rho"]
                                >= max_demand_g_s
                            ):  # m3/s * g/m3 = g/s
                                found_pc_large_enough = True

                    self.remove_dn0(new_pcs, is_there_always_mass_flow)
                    self._override_gas_pipe_classes[connected_asset] = new_pcs

                if asset in self.energy_system_components.get("heat_demand", []):
                    try:
                        max_demand = min(
                            max(self.get_timeseries(f"{asset}.target_heat_demand").values),
                            bounds[f"{asset}.Heat_demand"][1],
                        )
                    except KeyError:
                        max_demand = bounds[f"{asset}.Heat_demand"][1]
                    is_there_always_mass_flow = False
                    try:
                        if min(self.get_timeseries(f"{asset}.target_heat_demand").values) > 0.0:
                            is_there_always_mass_flow = True
                    except KeyError:
                        is_there_always_mass_flow = False

                    max_demand *= 1.3  # 30% added for expected worst case heat losses

                    __limit_list_available_heat_pipe_classes_max_power(max_demand, demand=True)

            # Here we do the same for sources as for the sources.
            for asset, (
                connected_asset,
                _orientation,
            ) in self.energy_system_topology.sources.items():
                if asset in self.energy_system_components.get("gas_source", []):
                    try:
                        max_prod_g_s = min(
                            max(self.get_timeseries(f"{asset}.maximum_gas_source").values),
                            bounds[f"{asset}.Gas_source_mass_flow"][1],
                        )
                    except KeyError:
                        max_prod_g_s = bounds[f"{asset}.Gas_source_mass_flow"][1]
                    new_pcs = []
                    found_pc_large_enough = False
                    for pc in self.gas_pipe_classes(connected_asset):
                        if not found_pc_large_enough:
                            new_pcs.append(pc)
                            if (
                                new_pcs[-1].maximum_discharge
                                * self.parameters(0)[f"{connected_asset}.rho"]
                                >= max_prod_g_s
                            ):
                                found_pc_large_enough = True
                    self._override_gas_pipe_classes[connected_asset] = new_pcs
                if asset in self.energy_system_components.get("heat_source", []):
                    try:
                        max_prod = min(
                            max(self.get_timeseries(f"{asset}.maximum_heat_source").values),
                            bounds[f"{asset}.Heat_source"][1],
                        )
                    except KeyError:
                        max_prod = bounds[f"{asset}.Heat_source"][1]
                    is_there_always_mass_flow = False
                    __limit_list_available_heat_pipe_classes_max_power(max_prod, demand=False)

        else:
            logger.warning("Limiting pipe classes do not cater for varying temperature yet")
        # ------------------------------------------------------------------------------------------

        for asset in [
            *self.energy_system_components.get("heat_source", []),
            *self.energy_system_components.get("ates", []),
            *self.energy_system_components.get("heat_buffer", []),
            *self.energy_system_components.get("heat_pump", []),
        ]:
            esdl_asset = self.esdl_assets[self.esdl_asset_name_to_id_map[asset]]
            for constraint in esdl_asset.attributes.get("constraint", []):
                if constraint.name == "setpointconstraint":
                    time_unit = constraint.range.profileQuantityAndUnit.perTimeUnit
                    if time_unit == esdl.TimeUnitEnum.HOUR:
                        time_hours = 1
                    elif time_unit == esdl.TimeUnitEnum.DAY:
                        time_hours = 24
                    elif time_unit == esdl.TimeUnitEnum.WEEK:
                        time_hours = 24 * 7
                    elif time_unit == esdl.TimeUnitEnum.MOTH:
                        time_hours = 24 * 7 * 30
                    elif time_unit == esdl.TimeUnitEnum.DAY:
                        time_hours = 365 * 24
                    else:
                        logger.error(
                            f"{asset} has a setpoint constaint specified with unknown"
                            f"per time unit"
                        )

                    asset_setpoints = {asset: (time_hours, constraint.range.maxValue)}

                    self._timed_setpoints.update(asset_setpoints)

    def remove_dn0(self, new_pcs: List[PipeClass], is_there_always_mass_flow: bool) -> None:
        """Remove pipe DN0 from the available pipe list, if there is always flow required"""
        # TODO: Bug to be resolved. Currently the solution is infeasible when only
        # 1 pipe class is available, but it should be able to working.
        # Then remove the need for having 2 pipes classes available in new_pcs.
        if is_there_always_mass_flow and len(new_pcs) > 2 and new_pcs[0].maximum_discharge == 0.0:
            new_pcs.remove(new_pcs[0])

    def pipe_classes(self, p):
        return self._override_pipe_classes.get(p, [])

    def gas_pipe_classes(self, p):
        return self._override_gas_pipe_classes.get(p, [])

    def esdl_heat_model_options(self):
        """Overwrites the fraction of the minimum tank volume"""
        options = super().esdl_heat_model_options()
        options["min_fraction_tank_volume"] = 0.0
        return options

    def temperature_carriers(self):
        return self.esdl_carriers_typed(type=[str(NetworkSettings.NETWORK_TYPE_HEAT).lower()])

    def electricity_carriers(self):
        return self.esdl_carriers_typed(
            type=[str(NetworkSettings.NETWORK_TYPE_ELECTRICITY).lower()]
        )

    def gas_carriers(self):
        return self.esdl_carriers_typed(
            type=[
                str(NetworkSettings.NETWORK_TYPE_GAS).lower(),
                str(NetworkSettings.NETWORK_TYPE_HYDROGEN).lower(),
            ]
        )

    def temperature_regimes(self, carrier):
        temperature_options = []
        temperature_step = 2.5

        try:
            temperature_options = self.__temperature_options[carrier]
        except KeyError:
            for asset in [
                *self.energy_system_components.get("heat_source", []),
                *self.energy_system_components.get("ates", []),
                *self.energy_system_components.get("heat_buffer", []),
                *self.energy_system_components.get("heat_pump", []),
                *self.energy_system_components.get("heat_exchanger", []),
                *self.energy_system_components.get("heat_demand", []),
            ]:
                esdl_asset = self.esdl_assets[self.esdl_asset_name_to_id_map[asset]]
                parameters = self.parameters(0)
                for i in range(len(esdl_asset.attributes["constraint"].items)):
                    constraint = esdl_asset.attributes["constraint"].items[i]
                    if (
                        constraint.name == "supply_temperature"
                        and carrier == parameters[f"{asset}.T_supply_id"]
                    ) or (
                        constraint.name == "return_temperature"
                        and carrier == parameters[f"{asset}.T_return_id"]
                    ):
                        try:
                            lb = self.__temperature_options[carrier][0]
                            ub = self.__temperature_options[carrier][-1]
                            lb = constraint.range.minValue if constraint.range.minValue > lb else lb
                            ub = constraint.range.maxValue if constraint.range.maxValue < ub else ub
                        except KeyError:
                            lb = constraint.range.minValue
                            ub = constraint.range.maxValue
                        n_options = round((ub - lb) / temperature_step)
                        temperature_options = np.linspace(lb, ub, n_options + 1)
                        if (
                            constraint.range.profileQuantityAndUnit.unit
                            != esdl.UnitEnum.DEGREES_CELSIUS
                        ):
                            RuntimeError(
                                f"{asset} has a temperature constraint with wrong unit "
                                f"{constraint.range.profileQuantityAndUnit.unit}, should "
                                f"always be in degrees celcius."
                            )
                        self.__temperature_options[carrier] = temperature_options

        return temperature_options



import logging
import unittest
import unittest.mock
from pathlib import Path
from unittest import TestCase

from mesido.esdl.esdl_parser import ESDLFileParser
from mesido.esdl.profile_parser import ProfileReaderFromFile
from mesido.exceptions import MesidoAssetIssueError
from mesido.potential_errors import MesidoAssetIssueType, PotentialErrors
from mesido.util import run_esdl_mesido_optimization
from mesido.workflows import EndScenarioSizing
from mesido.workflows.utils.error_types import mesido_issue_type_gen_message

from models.cost_information_errors import test_check_cost_information

import numpy as np

logger = logging.getLogger("mesido")
logger.setLevel(logging.INFO)


class TestCostInputCheck(TestCase):
    # NOTE: These tests are temporarily disabled because cost attribute errors
    # (ASSET_COST_ATTRIBUTE_INCORRECT and ASSET_COST_ATTRIBUTE_MISSING) have been
    # temporarily excluded from HEAT_NETWORK_ERRORS to generate warnings instead of errors.
    # TODO: Re-enable these tests when cost errors are added back to HEAT_NETWORK_ERRORS

    @unittest.skip(
        "Temporarily disabled: ASSET_COST_ATTRIBUTE_INCORRECT and "
        "ASSET_COST_ATTRIBUTE_MISSING have been excluded from HEAT_NETWORK_ERRORS to "
        "generate warnings instead of raising exceptions. Re-enable when cost errors "
        "are added back to HEAT_NETWORK_ERRORS."
    )
    def test_incorrect_cost_fail(self):
        """
        Test that ESDL files with incorrect cost attributes raise appropriate errors.

        This test verifies that the error checking mechanism correctly identifies and raises
        an exception when an ESDL file contains assets with incorrect cost attribute specifications.
        Specifically, it tests a case where an asset has installation costs specified per unit WATT
        when it should be None.

        The test uses a mock to replace the global POTENTIAL_ERRORS singleton with a fresh instance,
        ensuring that previous errors don't affect this test.

        Checks:
        1. The correct error type (ASSET_COST_ATTRIBUTE_INCORRECT) is raised
        2. The error message correctly identifies the general issue
        3. The error contains the specific asset ID and a detailed message explaining the problem
        """

        base_folder = Path(test_check_cost_information.__file__).resolve().parent

        with (
            self.assertRaises(MesidoAssetIssueError) as cm,
            unittest.mock.patch("mesido.potential_errors.POTENTIAL_ERRORS", PotentialErrors()),
        ):
            _ = run_esdl_mesido_optimization(
                EndScenarioSizing,
                base_folder=base_folder,
                esdl_file_name="graph_HDemands_incl_demand_4_incorrect_no_db.esdl",
                esdl_parser=ESDLFileParser,
                profile_reader=ProfileReaderFromFile,
                input_timeseries_file="profiles.csv",
            )

        # Check that the cold demand had an error
        np.testing.assert_equal(
            cm.exception.error_type, MesidoAssetIssueType.ASSET_COST_ATTRIBUTE_INCORRECT
        )
        np.testing.assert_equal(
            cm.exception.general_issue,
            mesido_issue_type_gen_message(MesidoAssetIssueType.ASSET_COST_ATTRIBUTE_INCORRECT),
        )
        np.testing.assert_equal(
            cm.exception.message_per_asset_id["f4fd7ca1-3f10-4bdb-88b3-daf8c456d959"],
            "Specified installation costs of asset Producer_2 include a component per unit WATT,"
            " but should be None.",
        )

    @unittest.skip(
        "Temporarily disabled: ASSET_COST_ATTRIBUTE_MISSING has been excluded "
        "from HEAT_NETWORK_ERRORS to generate warnings instead of raising "
        "exceptions. Re-enable when cost errors are added back."
    )
    def test_missing_cost_fail(self):
        """
        Test that ESDL files with missing cost attributes raise appropriate errors.

        This test verifies that the error checking mechanism correctly identifies and raises
        an exception when an ESDL file contains assets with missing cost attributes that are
        required. Specifically, it tests a case where an asset has investment costs, which
        are required.

        The test uses a mock to replace the global POTENTIAL_ERRORS singleton with a fresh instance,
        ensuring that previous errors don't affect this test.

        Checks:
        1. The correct error type (ASSET_COST_ATTRIBUTE_MISSING) is raised
        2. The error message correctly identifies the general issue
        3. The error contains the specific asset ID and a detailed message explaining the problem
        """

        # Test missing investment cost, which is required for HeatProducer

        base_folder = Path(test_check_cost_information.__file__).resolve().parent

        with (
            self.assertRaises(MesidoAssetIssueError) as cm,
            unittest.mock.patch("mesido.potential_errors.POTENTIAL_ERRORS", PotentialErrors()),
        ):
            _ = run_esdl_mesido_optimization(
                EndScenarioSizing,
                base_folder=base_folder,
                esdl_file_name="graph_HDemands_incl_demand_4_missing_investment_no_db.esdl",
                esdl_parser=ESDLFileParser,
                profile_reader=ProfileReaderFromFile,
                input_timeseries_file="profiles.csv",
            )

        # Check that the cold demand had an error
        np.testing.assert_equal(
            cm.exception.error_type, MesidoAssetIssueType.ASSET_COST_ATTRIBUTE_MISSING
        )
        np.testing.assert_equal(
            cm.exception.general_issue,
            mesido_issue_type_gen_message(MesidoAssetIssueType.ASSET_COST_ATTRIBUTE_MISSING),
        )
        np.testing.assert_equal(
            cm.exception.message_per_asset_id["f4fd7ca1-3f10-4bdb-88b3-daf8c456d959"],
            "No investment costs information specified for Producer_2 of type HeatProducer.",
        )

        # Test missing operational cost, which is required for HeatProducer

        with (
            self.assertRaises(MesidoAssetIssueError) as cm,
            unittest.mock.patch("mesido.potential_errors.POTENTIAL_ERRORS", PotentialErrors()),
        ):
            _ = run_esdl_mesido_optimization(
                EndScenarioSizing,
                base_folder=base_folder,
                esdl_file_name="graph_HDemands_incl_demand_4_missing_operational_no_db.esdl",
                esdl_parser=ESDLFileParser,
                profile_reader=ProfileReaderFromFile,
                input_timeseries_file="profiles.csv",
            )

        # Check that the cost error was raised
        np.testing.assert_equal(
            cm.exception.error_type, MesidoAssetIssueType.ASSET_COST_ATTRIBUTE_MISSING
        )
        np.testing.assert_equal(
            cm.exception.general_issue,
            mesido_issue_type_gen_message(MesidoAssetIssueType.ASSET_COST_ATTRIBUTE_MISSING),
        )
        np.testing.assert_equal(
            cm.exception.message_per_asset_id["f4fd7ca1-3f10-4bdb-88b3-daf8c456d959"],
            (
                "No variable operational costs information specified for Producer_2 "
                "of type HeatProducer."
            ),
        )


if __name__ == "__main__":
    test_cold_demand = TestCostInputCheck()
    test_cold_demand.test_incorrect_cost_fail()
    test_cold_demand.test_missing_cost_fail()



from .airwater_heat_pump_elec import AirWaterHeatPumpElec
from .elec_heat_source_elec import ElecHeatSourceElec
from .electrolyzer import Electrolyzer
from .gas_heat_source_gas import GasHeatSourceGas

__all__ = [
    "AirWaterHeatPumpElec",
    "ElecHeatSourceElec",
    "Electrolyzer",
    "GasHeatSourceGas",
]



from mesido.pycml.component_library.milp.gas.gas_base import GasPort
from mesido.pycml.component_library.milp.heat.heat_source_gas import HeatSourceGas
from mesido.pycml.pycml_mixin import add_variables_documentation_automatically


@add_variables_documentation_automatically
class GasHeatSourceGas(HeatSourceGas):
    """
    The gas heat source gas component is used to model the source behaviour of gas boilers
    with gas input port. It functions as a heat source gas that supplies thermal energy
    to the connected heat network. Besides, it models the gas consumed by the asset.

    Variables created:
        {add_variable_names_for_documentation_here}

    Parameters:
        name : The name of the asset. \n
        modifiers : Dictionary with asset information.


    """

    def __init__(self, name, **modifiers):
        super().__init__(
            name,
            **modifiers,
        )
        self.component_subtype = "gas_heat_source_gas"

        self.id_mapping_carrier = -1

        # Assumption: heat in/out and added is nonnegative
        # Heat in the return (i.e. cold) line is zero
        self.add_variable(GasPort, "GasIn")

        self.add_equation(
            (
                (self.GasIn.mass_flow - self.Gas_demand_mass_flow)
                / (self.Q_nominal_gas * self.density)
            )
        )
